#include <stdint.h>

#include <avr/io.h>
#include <avr/interrupt.h>

#include "ir.h"
#include "pc_commands.h"
#include "button.h"
#include "led.h"

volatile uint8_t button_held_count;

// 4ms timer.
ISR(TIM0_COMPA_vect) {
	// USB keyboard idle timer.
	usb_keyboard_idle();
	// Update the LED output.
	led_update();
	// Handle the button.
	if (!(BUTTON_PIN & _BV(BUTTON_BIT))) {
		if (button_held_count < 0xFF) ++button_held_count;
	} else {
		button_held_count = 0;
	}
}

// Called when an NEC IR code has been received. The code is then translated into keystrokes and sent via the USB keyboard code.
void ir_nec_code_received(uint8_t address, uint8_t command, uint8_t repeat_count) {
	
	// Only address 0x00 is supported.
	if (address != 0x00) return;
	
	enum pc_command_t converted_command = PC_COMMAND_NONE;
	
	switch (command) {
		// Power.
		case 0x0C: converted_command = PC_COMMAND_POWER; break;
		// Exit.
		case 0x45: converted_command = PC_COMMAND_CANCEL; break;
		// Shift+exit
		case 0x59: converted_command = PC_COMMAND_CANCEL; break;
		// Menu.
		case 0x49: converted_command = PC_COMMAND_MENU_AUTO; break;
		// Shift+menu
		case 0x0F: converted_command = PC_COMMAND_TOGGLE_MODE; break;
		// Directional navigation.
		case 0x44: converted_command = PC_COMMAND_UP; break;
		case 0x48: converted_command = PC_COMMAND_DOWN; break;
		case 0x4C: converted_command = PC_COMMAND_LEFT; break;
		case 0x40: converted_command = PC_COMMAND_RIGHT; break;
		// OK.
		case 0x06: converted_command = PC_COMMAND_ACCEPT; break;
		// A/V source 1 (angle).
		case 0x54: converted_command = PC_COMMAND_NEXT_ANGLE; break;
		// A/V source 2 (zoom).
		case 0x03: converted_command = PC_COMMAND_TOGGLE_FULL_SCREEN; break;
		// I-II (audio language).
		case 0x5A: converted_command = PC_COMMAND_NEXT_AUDIO_STREAM; break;
		// Stereo (subtitles).
		case 0x4E: converted_command = PC_COMMAND_NEXT_SUBTITLE; break;
		// Channel keys.
		case 0x10: converted_command = PC_COMMAND_FORWARD; break;
		case 0x14: converted_command = PC_COMMAND_REVERSE; break;
		// Volume keys.
		case 0x50: converted_command = PC_COMMAND_VOLUME_UP; break;
		case 0x51: converted_command = PC_COMMAND_VOLUME_DOWN; break;
		// Mute.
		case 0x08: converted_command = PC_COMMAND_MUTE; break;
		// Shift+mute.
		case 0x4D: converted_command = PC_COMMAND_DISPLAY; break;
		// Rewind.
		case 0x1C: converted_command = USB_KEYBOARD_KEY_P; break;
		// Play.
		case 0x57: converted_command = PC_COMMAND_PLAY; break;
		// Fast forward.
		case 0x18: converted_command = PC_COMMAND_FORWARD; break;
		// Stop.
		case 0x5B: converted_command = PC_COMMAND_STOP; break;
		// Pause.
		case 0x5F: converted_command = PC_COMMAND_PAUSE; break;
		// TV guide.
		case 0x0D: converted_command = PC_COMMAND_OPEN; break;
		// Numeric keys.
		case 0x09: converted_command = PC_COMMAND_1; break;
		case 0x05: converted_command = PC_COMMAND_2; break;
		case 0x01: converted_command = PC_COMMAND_3; break;
		case 0x4B: converted_command = PC_COMMAND_4; break;
		case 0x47: converted_command = PC_COMMAND_5; break;
		case 0x43: converted_command = PC_COMMAND_6; break;
		case 0x4A: converted_command = PC_COMMAND_7; break;
		case 0x46: converted_command = PC_COMMAND_8; break;
		case 0x42: converted_command = PC_COMMAND_9; break;
		case 0x41: converted_command = PC_COMMAND_0; break;
		// Coloured keys.
		case 0x5D: converted_command = PC_COMMAND_RED; break;
		case 0x0B: converted_command = PC_COMMAND_GREEN; break;
		case 0x52: converted_command = PC_COMMAND_YELLOW; break;
		case 0x56: converted_command = PC_COMMAND_BLUE; break;
	}
	
	// Send the command.
	pc_command_send_repeating(converted_command, repeat_count);
}

// Called when a SIRCS IR code is received.
void ir_sircs_code_received(uint8_t address, uint8_t command, uint8_t repeat_count) {
	
	enum pc_command_t converted_command = PC_COMMAND_NONE;
	
	// Identify the pressed button.
	
	// General buttons.
	switch (command) {
		case 0x00: converted_command = PC_COMMAND_1; break;
		case 0x01: converted_command = PC_COMMAND_2; break;
		case 0x02: converted_command = PC_COMMAND_3; break;
		case 0x03: converted_command = PC_COMMAND_4; break;
		case 0x04: converted_command = PC_COMMAND_5; break;
		case 0x05: converted_command = PC_COMMAND_6; break;
		case 0x06: converted_command = PC_COMMAND_7; break;
		case 0x07: converted_command = PC_COMMAND_8; break;
		case 0x08: converted_command = PC_COMMAND_9; break;
		case 0x09: converted_command = PC_COMMAND_0; break;
		case 0x0B: converted_command = PC_COMMAND_ACCEPT; break;
		case 0x10: converted_command = PC_COMMAND_NEXT; break;
		case 0x11: converted_command = PC_COMMAND_PREVIOUS; break;
		case 0x12: converted_command = PC_COMMAND_VOLUME_UP; break;
		case 0x13: converted_command = PC_COMMAND_VOLUME_DOWN; break;
		case 0x14: converted_command = PC_COMMAND_MUTE; break;
		case 0x15: converted_command = PC_COMMAND_POWER; break;
	}
	
	switch (address) {
		case 0x03:
			// Teletext
			switch (command) {
				case 0x4C: converted_command = PC_COMMAND_RED; break;
				case 0x4D: converted_command = PC_COMMAND_GREEN; break;
				case 0x4E: converted_command = PC_COMMAND_YELLOW; break;
				case 0x4F: converted_command = PC_COMMAND_BLUE; break;
			}
			break;
		case 0x1A:
			// DVD player.
			switch (command) {
				case 0x0E: converted_command = PC_COMMAND_CANCEL; break;
				case 0x1A: converted_command = PC_COMMAND_MENUS; break;
				case 0x1B: converted_command = PC_COMMAND_ROOT_MENU_AUTO; break;
				case 0x2A: converted_command = PC_COMMAND_REPEAT_RANGE; break;
				case 0x2C: converted_command = PC_COMMAND_REPEAT; break;
				case 0x30: converted_command = PC_COMMAND_PREVIOUS; break;
				case 0x31: converted_command = PC_COMMAND_NEXT; break;
				case 0x32: converted_command = PC_COMMAND_PLAY; break;
				case 0x33: converted_command = PC_COMMAND_REVERSE; break;
				case 0x34: converted_command = PC_COMMAND_FORWARD; break;
				case 0x35: converted_command = PC_COMMAND_SHUFFLE; break;
				case 0x38: converted_command = PC_COMMAND_STOP; break;
				case 0x39: converted_command = PC_COMMAND_PAUSE; break;
				case 0x54: converted_command = PC_COMMAND_DISPLAY; break;
				case 0x60: converted_command = PC_COMMAND_STEP_BACKWARD; break;
				case 0x61: converted_command = PC_COMMAND_STEP_FORWARD; break;
				case 0x63: converted_command = PC_COMMAND_NEXT_SUBTITLE; break;
				case 0x64: converted_command = PC_COMMAND_NEXT_AUDIO_STREAM; break;
				case 0x65: converted_command = PC_COMMAND_NEXT_ANGLE; break;
				case 0x79: converted_command = PC_COMMAND_UP; break;
				case 0x7A: converted_command = PC_COMMAND_DOWN; break;
				case 0x7B: converted_command = PC_COMMAND_LEFT; break;
				case 0x7C: converted_command = PC_COMMAND_RIGHT; break;
				// PS2
				case 0x5F: converted_command = PC_COMMAND_RED; break;
				case 0x5C: converted_command = PC_COMMAND_GREEN; break;
				case 0x5D: converted_command = PC_COMMAND_YELLOW; break;
				case 0x5E: converted_command = PC_COMMAND_BLUE; break;
				case 0x50: converted_command = PC_COMMAND_TOGGLE_MODE; break;
			}
			break;
	}
	
	// Send the command.
	pc_command_send_repeating(converted_command, repeat_count);
}

int main(void) {
	
	// Initialise the LED.
	led_init();
	
	// Initialise the IR decoder.
	ir_init();
	
	// Initialise the USB keyboard handler.
	usb_keyboard_init();

	// Initialise timer 0 to use it as a 4ms timer.
	TCCR0A |= _BV(WGM01); // Mode = 2, CTC.
	TCCR0B |= _BV(CS02); // clkIO/256
	OCR0A = ((256 * 4) / ((1000.0 * 256 * 256) / F_CPU)) - 1; // Desired interrupt frequency is once every 4ms.
	TIMSK0 |= _BV(OCIE0A);
	
	// Initialise the input button.
	BUTTON_PORT |= _BV(BUTTON_BIT); // Enable the pull-up.
	
	// Enable interrupts.
	sei();
	
	// Flash the LED to indicate we're running.
	led_blink(LED_BLINK_PATTERN_3_MEDIUM);
	
	// Infinite loop.
	for(;;) {
		// USB polling.
		usb_keyboard_poll();
		// Is the button pressed?
		if (button_held_count == 10) {
			++button_held_count; // So we don't accidentally trigger this again.
			pc_command_send(PC_COMMAND_TOGGLE_MODE);
		}
	}

}